// Project 53 - An AVR-based LCD Digital Clock

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>

// variabes to store time and date
uint8_t hours, minutes, seconds, dow, dom, mo, years;	

void I2Cenable()
// enable I2C bus
{
    TWBR = 72; // 100 kHz I2C bus
    TWCR |= (1 << TWEN); // enable I2C on PORTC4 and 5
}

void I2Cwait()
// wait until I2C finishes an operation
{
    // wait until bit TWINT in TWCR is set to 1
    while (!(TWCR & (1<<TWINT))); 
}

void I2CstartWait(unsigned char address)
{
    uint8_t status;
    while (1)
    {
        // send START condition
        TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
    
    	// wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status != 0b00001000) && (status != 0b00010000)) continue;
    
    	// send device address
    	TWDR = address;
    	TWCR = (1<<TWINT) | (1<<TWEN);
    
    	// wait until transmission completed
    	I2Cwait();
    
    	// check value of TWSR, and mask out status bits
    	status = TWSR & 0b11111000;
    	if ((status == 0b00100000 )||(status == 0b01011000)) 
    	{    	    
            // secondary device is busy, so send stop condition to terminate write operation
            TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
	        
            // wait until stop condition is executed and I2C bus released
            while(TWCR & (1<<TWSTO));
            continue;
        }
        break;
    }
}

void I2Cstop()
// stop I2C bus and release GPIO pins
{
    // clear interrupt, enable I2C, generate stop condition
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWSTO);
}

void I2Cwrite(uint8_t data)
// send 'data' to I2C bus
{
    TWDR = data;
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
}

uint8_t I2Cread()
// read incoming byte of data from I2C bus
{
    TWCR |= (1 << TWINT)|(1 << TWEN);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

uint8_t I2CreadACK()
// read incoming byte of data from I2C bus and ACK signal
{
    TWCR |= (1 << TWINT)|(1 << TWEN)|(1 << TWEA);
    I2Cwait();
    // incoming byte is placed in TWDR register
    return TWDR;	
}

uint8_t decimalToBcd(uint8_t val)
// Convert integer to BCD
{
    return((val/10*16)+(val%10));
}

uint8_t bcdToDec(uint8_t val)
// Convert BCD to integer
{
    return((val/16*10)+(val%16));
}

void setTimeDS3231(uint8_t hh, uint8_t mm, uint8_t ss, uint8_t dw, uint8_t dd, uint8_t mo, uint8_t yy)
// set the time on DS3231
{
    I2CstartWait(0xD0); // DS3231 write
    I2Cwrite(0x00);		// start with hours register
    I2Cwrite(decimalToBcd(ss));  // seconds
    I2Cwrite(decimalToBcd(mm));  // minutes
    I2Cwrite(decimalToBcd(hh));  // hours
    I2Cwrite(decimalToBcd(dw));  // day of week
    I2Cwrite(decimalToBcd(dd));  // date
    I2Cwrite(decimalToBcd(mo));	 // month
    I2Cwrite(decimalToBcd(yy));	 // year
    I2Cstop();	
}

void readTimeDS3231()
// retrieve time and date from DS3231
{
    I2CstartWait(0xD0); // DS3231 write
    I2Cwrite(0x00);		// seconds register
    I2CstartWait(0xD1); // DS3231 read
    seconds = bcdToDec(I2CreadACK());	
    minutes = bcdToDec(I2CreadACK());	
    hours = bcdToDec(I2CreadACK());	
    dow = bcdToDec(I2CreadACK());	
    dom = bcdToDec(I2CreadACK());
    mo = bcdToDec(I2CreadACK());
    years = bcdToDec(I2CreadACK());
}

void commandLCD(uint8_t _command )	
{
    // takes command byte and sends upper nibble, then lower nibble to LCD
    PORTD = (PORTD & 0x0F) | (_command & 0xF0); 
    PORTD &= ~(1<<PD0);	
    PORTD |= (1<<PD1);	
    _delay_us(1);			
    PORTD &= ~(1<<PD1);	
    _delay_us(200);		
    PORTD = (PORTD & 0x0F) | (_command << 4);	
    PORTD |= (1<<PD1);		
    _delay_us(1);
    PORTD &= ~(1<<PD1);	
    _delay_ms(2);
}

void initLCD()
{
    DDRD = 0b11111111;  
    _delay_ms(100);		
	
    commandLCD(0x02);	
    commandLCD(0x28);   
    commandLCD(0x0C);  
    commandLCD(0x06);  
    commandLCD(0x01); 

    _delay_ms(2);	
}

void clearLCD()			
{
    commandLCD (0x01);		
    _delay_ms(2);		
    commandLCD (0x80);	
    _delay_ms(2);		
}

void printLCD(char *_string)	
{
    uint8_t i;
    for(i=0; _string[i]!=0; i++)
    {
        PORTD = (PORTD & 0x0F) | (_string[i] & 0xF0);
        PORTD |= (1<<PD0);
        PORTD |= (1<<PD1);
        _delay_us(1);
        PORTD &= ~(1<<PD1);
        _delay_us(200);
        PORTD = (PORTD & 0x0F) | (_string[i] << 4);
        PORTD |= (1<<PD1);
        _delay_us(1);
        PORTD &= ~(1<<PD1);
        _delay_ms(2);
    }
}

void cursorLCD(uint8_t column, uint8_t row) 	
// move cursor to column (0~15), row (0~1) 
{
    if (row == 0 && column<16)
    {
        commandLCD((column & 0x0F)|0x80); 
    }			
    else if (row == 1 && column<16)
    {
        commandLCD((column & 0x0F)|0xC0);
    }			
}

int main()
{
    initLCD();
    I2Cenable();
    char numbers[9];
			
    // uncomment to set time and date, then comment and re-flash code
    // setTimeDS3231(8,50,0,3,16,6,21); // h,m,s,dow,dom,m,y
    while(1) 
    {
        readTimeDS3231();
	
        itoa(hours,numbers,10); // hours
        cursorLCD(4,0);		
        if (hours==0)
        {
            printLCD("00");
        } else if (hours>0 && hours <10)
        {
            printLCD("0");
            printLCD(numbers);
        } else if (hours>=10)
        {
        printLCD(numbers);
    }
		
    cursorLCD(6,0);
    printLCD(":");
		
    itoa(minutes,numbers,10); // minutes
    cursorLCD(7,0);
    if (minutes==0)
    {
        printLCD("00");
    } else if (minutes>0 && minutes <10)
    {
        printLCD("0");
        printLCD(numbers);
        } else if (minutes>=10)
        {
            printLCD(numbers);
        }
		
        cursorLCD(0,9);
        printLCD(":");
		
        itoa(seconds,numbers,10); // minutes
        cursorLCD(10,0);
        if (seconds==0)
        {
            printLCD("00");
        } else if (seconds>0 && seconds <10)
        {
            printLCD("0");
            printLCD(numbers);
        } else if (seconds>=10)
        {
            printLCD(numbers);
        }
		
        // day of week
        cursorLCD(2,1);
        switch(dow)
        {
            case 1 : printLCD("Mon"); break;
            case 2 : printLCD("Tue"); break;
            case 3 : printLCD("Wed"); break;
            case 4 : printLCD("Thu"); break;
            case 5 : printLCD("Fri"); break;
            case 6 : printLCD("Sat"); break;
            case 7 : printLCD("Sun"); break;
        }

        itoa(dom,numbers,10); // day of month
        cursorLCD(6,1);
        if (dom<10)
        {
            printLCD("0");
        }
        printLCD(numbers);
		
        cursorLCD(8,1);
        printLCD("/");		
		
        itoa(mo,numbers,10); // month
        cursorLCD(9,1);
        if (mo<10)
        {
            printLCD("0");
        }
        printLCD(numbers);
		
        cursorLCD(11,1);
        printLCD("/");		
		
        itoa(years,numbers,10); // year
        cursorLCD(12,1);
        printLCD(numbers);
	
        _delay_ms(900);		
        clearLCD(); // refresh LCD
    }
}